#ifndef __JABYENGINE_SPU_PORTS_HPP__
#define __JABYENGINE_SPU_PORTS_HPP__
#include "../Auxiliary/io_class_helper.hpp"

namespace SPU {
    //SampleRate is defined as 4096 == 44100Hz
    struct __no_align SampleRate {
        uint16_t raw_value = 0;

        static constexpr SampleRate from_HZ(long double freq)
        {
            constexpr long double Base = (4096.0 / 44100.0);

            return static_cast<SampleRate>(((freq*4096.0)/44100.0));
        }
    };

    struct __no_align SweepVolume {
        enum Type {
            Linear      = 0,
            Exponential = 1,
        };

        enum Direction {
            Increase = 0,
            Decrease = 1,
        };

        enum Phase {
            Positive = 0,
            Negative = 1,
        };

        uint16_t raw_value = 0;

        constexpr SweepVolume() = default;

        constexpr void set_volume(int16_t volume) {
            this->raw_value = bit::value::set_normalized(this->raw_value, bit::cast<uint16_t>(volume >> 1), __start_end_bit2_start_length(0, 14));
        }

        constexpr int16_t get_volume() const {
            return bit::cast<int16_t>(bit::value::get_normalized(this->raw_value, __start_end_bit2_start_length(0, 14)));
        }

        io_class__2option_map(volume_mode, sweep_mode, 15);
        io_class__2option_map_getter(Type,      linear_sweep_mode,    exponential_sweep_mode, sweep_mode_type,      14);
        io_class__2option_map_getter(Direction, increase_sweep_mode,  decrease_sweep_mode,    sweep_mode_direction, 13);
        io_class__2option_map_getter(Phase,     positive_sweep_phase, negative_sweep_phase,   sweep_phase,          12);

        //Uses only 5bit of shift (0..1F (slow..fast))
        constexpr void set_sweep_shift(uint8_5b shift) {
            const uint16_t crop_value = static_cast<uint16_t>(shift); //(bit::value::crop_value(shift, 5));
            this->raw_value = bit::value::set_normalized(this->raw_value, crop_value, __start_end_bit2_start_length(2, 6));
        }

        constexpr uint8_5b get_sweep_shift() const {
            return bit::value::get_normalized(this->raw_value, __start_end_bit2_start_length(2, 6));
        }
        
        //0..3 maps to => +7, +6, +5, +4 or -8, -7, -6, -5
        constexpr void set_sweep_step(uint8_2b step) {
            this->raw_value = bit::value::set_normalized(this->raw_value, static_cast<uint16_t>(step), __start_end_bit2_start_length(0, 1));
        }

        constexpr uint8_2b get_sweep_step() const {
            return bit::value::get_normalized(this->raw_value, __start_end_bit2_start_length(0, 1));
        }
    };

    struct __no_align Voice {
        SweepVolume volumeLeft;
        SweepVolume volumeRight;
        /*VolatileValue<SampleRate> sampleRate;
        VolatileValue<uint16_t>   adr;
        VolatileValue<uint16_t>   ad; //0x...08: 15 - 0
        VolatileValue<uint16_t>   sr; //0x...0A: 31 - 16
        VolatileValue<uint16_t>   currentVolume; //Not used
        VolatileValue<uint16_t>   repeatAdr;*/
    };
}

#endif //!__JABYENGINE_SPU_PORTS_HPP__